#include <stddef.h>
#include <stdlib.h>
#include <wchar.h>

#include "bigint.h"
#include "symtab.h"

static symtab *symtab_new(const wchar_t *name)
{
	symtab *p;

	p = malloc(sizeof(*p));
	if (p == NULL)
		return NULL;

	/* We take responsibility for a copy of name. */
	p->name = malloc(sizeof(wchar_t) * (wcslen(name) + 1));
	if (p->name == NULL) {
		free(p);
		return NULL;
	}
	wcscpy(p->name, name);

	return p;
}

symtab *symtab_set(symtab *sym, const wchar_t *name, const bigint *value)
{
	symtab *p;

	for (p = sym; p != NULL; p = p->next) {
		if (wcscmp(name, p->name) == 0) {
			bigint_free(&p->value);
			break;
		}
	}
	if (p == NULL) {
		p = symtab_new(name);
		if (p == NULL)
			return NULL;
		p->next = sym;
		sym = p;
	}
	if (!bigint_copy(&p->value, value)) {
		free(p->name);
		free(p);
	}

	return sym;
}

const bigint *symtab_lookup(const symtab *sym, const wchar_t *name)
{
	const symtab *p;

	for (p = sym; p != NULL; p = p->next) {
		if (wcscmp(name, p->name) == 0)
			return &p->value;
	}

	return NULL;
}

void symtab_free(symtab *sym)
{
	symtab *p, *next;

	for (p = sym; p != NULL; p = next) {
		next = p->next;
		free(p->name);
		bigint_free(&p->value);
		free(p);
	}
}
