#include <errno.h>
#include <stdio.h>
#include <wchar.h>

/* Read a single UTF-8–encoded character from fp. Returns WEOF on end-of-file or
   error. On error, sets errno; EILSEQ means there was an invalid or truncated
   multibyte sequence, or an illegal character. */
wint_t utf8_getc(FILE *fp)
{
	static const int N_BITS[6] = { 7, 11, 16, 21, 26, 31 };
	unsigned char prefix, mask;
	int c, i, n;
	wint_t wc;

	c = fgetc(fp);
	if (c == EOF)
		return WEOF;

	/* Special-case the one-byte encoding. */
	if ((c & 0x80) == 0x00)
		return (wint_t) c;

	prefix = 0xC0;
	mask = 0xE0;
	n = 1;
	while (mask != 0xFF) {
		if ((c & mask) == prefix)
			break;
		prefix = (prefix >> 1) | 0x80;
		mask = (mask >> 1) | 0x80;
		n++;
	}
	if (mask == 0xFF)
		goto badchar;

	wc = (wint_t) (c & ~mask) << (n * 6);
	for (i = 1; i < n + 1; i++) {
		errno = 0;
		c = fgetc(fp);
		if (c == EOF) {
			if (errno == 0)
				goto badchar;
			else
				return WEOF;
		}
		if ((c & 0xC0) != 0x80)
			goto badchar;
		wc |= (unsigned char) (c & ~0xC0) << (n - i) * 6;
	}

	/* Check for shortest encoding. */
	if (wc >> N_BITS[n - 1] == 0)
		goto badchar;

	/* Check for illegal characters. */
	if ((wc >= 0xD800 && wc <= 0xDFFF) || wc == 0xFFFE || wc == 0xFFFF)
		goto badchar;

	return wc;

badchar:
	errno = EILSEQ;
	return WEOF;
}
